#  filters.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import numpy as np
from scipy.signal import butter, filtfilt

from sensorialytics.mathematics.functions import Symmetrised, Sigmoid
from sensorialytics.mathematics.tools import derive

__all__ = ['Filter', 'ButterworthFilter', 'DerivativeFilter',
           'ComplementaryFilter', 'RecursiveAverageFilter']


class Filter:
    def __init__(self, sampling_frequency: float = 1.0):
        self._sampling_frequency = sampling_frequency
        self._order = 1
        self._initialized = False

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        """
        :param x: input signals
        :return: filtered signals
        """

        return x.copy()

    def set_sampling_frequency(self, sampling_frequency: float):
        self._sampling_frequency = sampling_frequency


class ButterworthFilter(Filter):
    def __init__(self, order: int, fn: float, filter_type: str,
                 fs: float = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._order = order
        self._fn = fn
        self._filter_type = filter_type

        self.set_sampling_frequency(sampling_frequency=fs)
        self._make_filter()

    def _make_filter(self):
        if self._sampling_frequency is not None:
            self.__B, self.__A = butter(N=self._order, Wn=self._fn,
                                        btype=self._filter_type,
                                        fs=self._sampling_frequency)
            self._initialized = True

    def set_sampling_frequency(self, sampling_frequency: float):
        super().set_sampling_frequency(sampling_frequency)

        self._make_filter()

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        """
        :param x: input signals
        :return: filtered signals
        """

        if self._initialized:
            return filtfilt(self.__B, self.__A, x)
        else:
            raise RuntimeError('Filter not initialized')


class ComplementaryFilter(Filter):
    def __init__(self, gain: float, *args, **kwargs):
        """
        :param gain : gain of the filter
        """
        super().__init__(*args, **kwargs)

        if not (0 < gain < 1):
            raise ValueError('Must be 0 < gain < 1')

        self._gain = gain
        self._complementary_gain = 1.0 - gain

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        filtered = []
        value = x[0]
        dx = np.array(args[0])

        for xx, dxx in zip(x, dx):
            value = self._gain * xx + self._complementary_gain * (value + dxx)
            filtered.append(value)

        return np.array(filtered)


class DCBlocker(Filter):
    def __init__(self, memory_gain: float, momentum_gain: float, *args,
                 **kwargs):
        """
        :param memory_gain : memory gain
        :param momentum_gain: momentum gain
        """
        super().__init__(*args, **kwargs)

        self._memory_gain = memory_gain
        self._momentum_gain = momentum_gain
        self._complementary_momentum_gain = (1 - self._momentum_gain)

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        """
        :param x: input signals
        :return: filtered signals
        """

        filtered = [0]
        y = 0
        dxx_old = 0
        dx = (x[1:] - x[:-1])

        for dxx in dx:
            y = self._memory_gain * y + \
                self._momentum_gain * dxx_old + \
                self._complementary_momentum_gain * dxx

            dxx_old = dxx
            filtered.append(y)

        return np.array(filtered)


class DerivativeFilter(Filter):
    def __init__(self, sigma: float = 1.0, cutoff_derivative: float = 0.0,
                 alpha: float = 1.0, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._dt = 1.0 / self._sampling_frequency
        self._weight_function = Symmetrised(
            function=Sigmoid(sigma=sigma, x0=cutoff_derivative)
        )

        if (0.0 <= alpha) and (alpha <= 1.0):
            self._alpha = alpha
        else:
            raise ValueError('alpha must be so that 0.0 <= alpha <= 1.0')

    def set_sampling_frequency(self, sampling_frequency: float):
        super().set_sampling_frequency(sampling_frequency)
        self._dt = 1.0 / self._sampling_frequency

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        derivative = derive(x=x, dx=self._dt)
        filtered = np.zeros(len(x))

        filtered[0] = x[0]

        for i, der in enumerate(derivative[1:]):
            weight = self._weight_function(der)
            delta = weight * der * self._dt
            value = filtered[i] + delta
            filtered[i + 1] = x[i] + self._alpha * (value - x[i])

        return filtered


class RecursiveAverageFilter(Filter):
    def __init__(self, sampling_frequency: float, gain: float):
        super().__init__(sampling_frequency=sampling_frequency)

        self._gain = gain
        self._complementary_gain = 1.0 - gain

    def filter(self, x: np.ndarray, *args) -> np.ndarray:
        filtered = np.zeros(len(x))
        y = x[0]

        for i, xx in enumerate(x):
            y = self._complementary_gain * y + self._gain * xx
            filtered[i] = y

        return filtered



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import constants
from . import functions
from . import point
from . import polynomial
from . import tools

from .constants import *  # NOQA
from .functions import *  # NOQA
from .point import *  # NOQA
from .polynomial import *  # NOQA
from .tools import *  # NOQA

__all__ = []

__all__ += constants.__all__
__all__ += functions.__all__
__all__ += point.__all__
__all__ += polynomial.__all__
__all__ += tools.__all__



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import datetime
import functools
import logging

__all__ = ['execution_time']


def execution_time(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        start = datetime.datetime.now()
        logging.info(f'Execution started on: {start}')
        res = func(*args, **kwargs)
        end = datetime.datetime.now()
        logging.info(f'Execution ended on: {end}')
        logging.info(f'Execution time: {end - start}')

        return res

    return wrapped



#  fitfile.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import fitdecode
import pandas as pd

__all__ = ['FitFile']

OTHER_FIT_FORMATS = (fitdecode.FitHeader, fitdecode.FitDefinitionMessage,
                     fitdecode.FitDataMessage, fitdecode.FitCRC)


class FitFile:
    def __init__(self, file_path: str):
        self._df = pd.DataFrame([])
        self._power = pd.Series([])
        self._recap = None

        self.__read(file_path=file_path)

    @property
    def df(self) -> pd.DataFrame:
        """
        :return: pandas.DataFrame = the whole DataFrame of the FIT file
        """

        return self._df

    @property
    def recap(self) -> dict:
        """
        :return: dict = a recap of the FIT file
        """

        return self._recap

    @property
    def power(self) -> pd.Series:
        """
        :return: pandas.Series = the power calculated during the session
        """
        return self._power

    def save_to_csv(self, csv_path: str) -> None:
        """
        :param csv_path: str = path for the output file
        """

        self._df.to_csv(path_or_buf=csv_path)

    def __read(self, file_path: str):
        frames = self.__read_fit(file_path=file_path)

        self.__get_df(frames=frames)
        self.__get_power()
        self.__get_recap()

    def __get_df(self, frames: list):
        self._df = pd.DataFrame({
            tick: {v.name: v.value for v in f.fields}
            for tick, f in enumerate(frames)
        }).T

    @staticmethod
    def __read_fit(file_path: str) -> list:
        frames = []
        with fitdecode.FitReader(file_path) as fit:
            for frame in fit:
                if isinstance(frame, fitdecode.FitDataMessage):
                    frames.append(frame)
                elif isinstance(frame, OTHER_FIT_FORMATS):
                    pass

            return frames

    def __get_power(self):
        valid_pow = ~self._df.power.isna()
        power = self._df.loc[valid_pow, ['timestamp', 'power']].copy()
        power.set_index('timestamp', inplace=True)

        self._power = power['power']

    def __get_recap(self):
        self._recap = dict(
            settings={},
            various={},
            data_columns=[],
            unknowns={},
            useless=[]
        )

        for c in self._df.columns:
            unique = [x for x in self._df[c].unique() if not pd.isna(x)]
            n_unique = len(unique)

            if c[0:8] == 'unknown_':
                self._recap['unknowns'].update({c: unique})
            else:
                if n_unique > 0:
                    if n_unique == 1:
                        self._recap['settings'].update({c: unique[0]})
                    elif n_unique < 10:
                        self._recap['various'].update({c: unique})
                    else:
                        self._recap['data_columns'].append(c)
                else:
                    self._recap['useless'].append(c)



#  detectors.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from typing import Tuple

import numpy as np
import pandas as pd

__all__ = ['Detector', 'Window', 'SymmetricWindow', 'PeaksValleysDetector']

F = 0.8

DetectorInput = (np.ndarray, pd.Series)


class Detector:
    def __init__(self):
        pass


class PeaksValleysDetector(Detector):
    def __init__(self):
        super().__init__()

        self._window = SymmetricWindow()
        self._start_window_width = 1

        self._sampling_frequency = 0.0

        self._min_height = 0.0

        self._max_reset_time = 2.0
        self._reset_time = 0.0

        self._peaks = np.array([]).reshape((-1, 2))
        self._valleys = np.array([]).reshape((-1, 2))
        self._mv_avg = []
        self._found_peak = False
        self._found_valley = False

    def detect(self, x: DetectorInput, start_sec_in_window: float,
               min_sec_in_window: float, max_sec_in_window: float,
               min_height: float,
               sampling_frequency: float) -> Tuple[np.ndarray, np.ndarray]:
        self.reset()

        self._sampling_frequency = sampling_frequency
        self._start_window_width = int(
            start_sec_in_window * sampling_frequency)
        self._min_height = min_height

        values = self.__get_values(x)

        self._window = SymmetricWindow(
            x=values,
            width=self._start_window_width,
            min_width=int(min_sec_in_window * sampling_frequency),
            max_width=int(max_sec_in_window * sampling_frequency)
        )

        self.__detect(values)

        return np.array(self._peaks), np.array(self._valleys)

    def reset(self):
        self._peaks = np.array([]).reshape((-1, 2))
        self._valleys = np.array([]).reshape((-1, 2))
        self._mv_avg = []
        self._found_peak = False
        self._found_valley = False

    @staticmethod
    def __get_values(x: DetectorInput) -> np.ndarray:
        if isinstance(x, pd.Series):
            return x.values
        else:
            return x

    def __detect(self, x: np.ndarray):
        for i in range(self._window.half_width, len(x)):
            x_range = self._window.get(i)

            self._mv_avg.append([i, x_range.mean()])

            self.__find_peak(i, x_range, x[i])
            self.__find_valley(i, x_range, x[i])
            self.__check_reset()

    def __find_peak(self, i, x_range, value):
        if self._found_valley:
            is_max = x_range.max() == value
            is_high_enough = value - self._mv_avg[-1][1] > self._min_height

            if is_max and is_high_enough:
                self._found_peak = True
                self._peaks = np.append(self._peaks, [[i, value]], axis=0)

                self.__on_found_all()

    def __find_valley(self, i, x_range, value):
        is_min = x_range.min() == value
        is_low_enough = self._mv_avg[-1][1] - value > self._min_height

        if is_min and is_low_enough:
            self._found_valley = True
            self._valleys = np.append(self._valleys, [[i, value]], axis=0)

    def __check_reset(self):
        if self._found_peak or self._found_valley:
            self._reset_time += 1.0 / self._sampling_frequency

            if self._reset_time > self._max_reset_time:
                self._reset_time = 0.0
                self._found_peak = False
                self._found_valley = False

                self._window.set_width(self._start_window_width)

    def __on_found_all(self):
        self._found_valley = False
        self._found_peak = False
        self._reset_time = 0.0

        delta = 2 * (self._valleys[-1, 0] - self._peaks[-1, 0])
        width = 0.8 * 0.5 * delta
        width = int(F * self._window.width + (1 - F) * width)

        self._window.set_width(width)

        self._max_reset_time = 2 * self._window.width / self._sampling_frequency


class Window:
    def __init__(self, x: np.ndarray = np.array([]),
                 width: int = 1, min_width: int = 0, max_width: int = np.inf):
        self._x = x
        self._width = 0
        self._min_width = min_width
        self._max_width = max_width

        self.set_width(width)

    @property
    def width(self):
        return self._width

    @property
    def min_width(self):
        return self._min_width

    @property
    def max_width(self):
        return self._max_width

    def set_width(self, width: int):
        self._width = max(self._min_width, min(width, self._max_width))

    def set_min_width(self, min_width: int):
        self._min_width = min_width

    def set_max_width(self, max_width: int):
        self._max_width = max_width


class SymmetricWindow(Window):
    def __init__(self, x: np.ndarray = np.array([]),
                 width: int = 3, max_width: int = 3, min_width: int = 3):
        self._half_width = 0

        super().__init__(x=x, width=width, min_width=min_width,
                         max_width=max_width)

    @property
    def half_width(self):
        return self._half_width

    def get(self, i: int) -> np.ndarray:
        start = max(0, i - self._half_width)
        end = min(i + self._half_width, len(self._x))

        return self._x[start:end]

    def set_width(self, width: int):
        super().set_width(width)
        self._half_width = self._width // 2



#  exceptions.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved


class FitError(Exception):
    def __init__(self, *args):
        super().__init__(*args)


class MaximumDepthError(Exception):
    def __init__(self, *args):
        super().__init__(*args)



#  polynomial.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from __future__ import annotations

from typing import Union

import numpy as np

__all__ = ['Polynomial']


class Polynomial:
    def __init__(self, *coefficients):
        self._coefficients = np.array([])
        self.__unpack_coefficients(*coefficients)
        self._degree = len(self._coefficients) - 1

        self._derivative = None
        self._integral = None

    @property
    def degree(self):
        return self._degree

    @property
    def coefficients(self):
        return self._coefficients

    @property
    def derivative(self) -> Polynomial:
        if self._derivative is None:
            coefficients = np.array([
                (n + 1) * c for n, c in enumerate(self._coefficients[1:])
            ])

            self._derivative = Polynomial(coefficients)

        return self._derivative

    @property
    def integral(self) -> Polynomial:
        if self._integral is None:
            coefficients = np.array([0.0] + [
                1.0 / (n + 1) * c for n, c in enumerate(self._coefficients)
            ])

            self._integral = Polynomial(coefficients)

        return self._integral

    def __str__(self):
        if self._degree >= 0:
            return str(self._coefficients[0]) + " + " + " + ".join([
                f'{k} * x^{n + 1}'
                for n, k in enumerate(self._coefficients[1:])
            ])
        else:
            return ''

    def __getitem__(self, item: int) -> float:
        return self._coefficients[item]

    def __setitem__(self, n: int, coefficient: float):
        if n > self._degree:
            self._coefficients = np.append(self._coefficients,
                                           np.zeros(n - self._degree))
            self._degree = n

        self._coefficients[n] = coefficient
        self._derivative = None
        self._integral = None

    def __call__(self, x: Union[float, np.ndarray]):
        return sum(c * x ** n for n, c in enumerate(self._coefficients))

    def __unpack_coefficients(self, *coefficients):
        for c in coefficients:
            if isinstance(c, (tuple, list, np.ndarray)):
                self._coefficients = np.append(self._coefficients, c)
            else:
                self._coefficients = np.append(self._coefficients, [c])



#  image.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import cv2
import numpy as np

__all__ = ['image_to_signal']


def image_to_signal(path: str, rescale=True):
    img = cv2.imread(path, 0)
    indices = np.array([np.arange(len(img)) + 1 for _ in img[0]]).T

    mask = (img.min(axis=0) < 150)

    signal = indices * (img < 150)
    signal = signal.sum(axis=0)[mask] / (signal > 0).sum(axis=0)[mask]

    if rescale:
        signal = (signal - signal.min()) / (signal.max() - signal.min())

    return signal



#  dates.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import datetime

import pandas

__all__ = ['get_age', 'get_date', 'get_date_ymd', 'get_date_y_m_d',
           'get_date_ymdhm']


def get_age(birth_date: (datetime.datetime, str),
            ref_date: (datetime.datetime, str)):
    if isinstance(birth_date, str) or pandas.isna(birth_date):
        birth_date = pandas.to_datetime(birth_date)

    if isinstance(ref_date, str) or pandas.isna(ref_date):
        ref_date = pandas.to_datetime(ref_date)

    corr = (ref_date.month, ref_date.day) < (birth_date.month, birth_date.day)
    age = ref_date.year - birth_date.year - corr

    return age


def get_date(n_days=0):
    now = datetime.datetime.now() + datetime.timedelta(n_days)
    date = {
        'ye': f'{now.year:4}',
        'mo': f'{now.month:2}',
        'da': f'{now.day:2}',
        'ho': f'{now.hour:2}',
        'mi': f'{now.minute:2}'
    }
    return date


def get_date_ymd(n_days=0):
    date = get_date(n_days)
    return date['ye'] + date['mo'] + date['da']


def get_date_y_m_d(n_days=0):
    date = get_date(n_days)
    return date['ye'] + '-' + date['mo'] + '-' + date['da']


def get_date_ymdhm(n_days=0):
    date = get_date(n_days)
    return date['ye'] + date['mo'] + date['da'] + '-' + date['ho'] + date['mi']



#  events.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import json

import pandas as pd

from . import helper as h

__all__ = ['Events']


class Events(pd.DataFrame):
    def __init__(self, events_path: str = None):
        """
        :param events_path: str = path of the events data csv
        """

        if events_path is None:
            super(Events, self).__init__(pd.DataFrame())
            return

        _, n_headers = h.read_header(events_path)

        super(Events, self).__init__(
            pd.read_csv(events_path, skiprows=n_headers))

        self[h.COL_LEADING_CORE_TICK] = self[h.COL_LEADING_CORE_TICK].astype(
            int)
        self.set_index(h.COL_LEADING_CORE_TICK, drop=True, inplace=True)

        if h.COL_EVENT_PARAMS in self.columns:
            self.__parse_events_parameters()
        else:
            self.__parse_events_column_old()

    @property
    def settings(self) -> dict:
        """
        :return: settings used during the session
        """

        return self.get_event(h.KEY_ON_SETTINGS_SAVED).iloc[0, :].to_dict()

    def get_event(self, event_name: str):
        """
        :param event_name: str = name of the event to get
        :return: pandas.DataFrame = DataFrame containing the all the occurrences
                                    of the selected event
        """

        condition = self[h.COL_EVENT_NAME] == event_name
        event = self[condition][h.COL_EVENT_PARAMS]

        index = event.index
        event = pd.DataFrame([p for p in event])
        event.index = index

        if len(event) == 0:
            return event

        event[h.COL_TIME] = self[h.COL_TIME].drop_duplicates()

        return event

    def __parse_events_parameters(self):
        def to_dict(x) -> dict:
            if x is None or pd.isna(x):
                return {}

            return json.loads(x)

        self[h.COL_EVENT_PARAMS] = self[h.COL_EVENT_PARAMS].apply(to_dict)

    def __parse_events_column_old(self):
        events = self[h.COL_EVENT].apply(
            lambda row: self.__take(row.split(':'), 0))

        events_info = self[h.COL_EVENT].apply(
            lambda row: self.__parse_event_info(row))

        self[h.COL_EVENT] = events
        self[h.COL_EVENT_INFO] = events_info

    def __parse_event_info(self, x):
        if x is None:
            return None

        all_info = self.__take(x.split(':'), 1)

        if all_info is None:
            return None

        parsed_event_info = {}

        for info in all_info.split('|'):
            info_kv = info.strip().split('=')
            k = self.__parse_value(info_kv[0].strip())
            v = self.__parse_value(info_kv[1].strip())

            parsed_event_info.update({k: v})

        return parsed_event_info

    @staticmethod
    def __take(x, i: int):
        try:
            return x[i].strip()
        except IndexError:
            return None

    @staticmethod
    def __parse_value(x):
        if x is None:
            return None

        if x in ['true', 'false', 'True', 'False']:
            return bool(x)

        try:
            return float(x)
        except ValueError:
            return x



#  core_data.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import logging
from typing import Tuple, AnyStr, Set, List, Dict

import numpy as np
import pandas as pd

from sensorialytics.helpers.tools import as_bool
from sensorialytics.mathematics.tools import cast_to_float
from sensorialytics.signals.filters import Filter
from .helper import get_session_id, read_header, FiltersType, \
    KEY_SAMPLING_FREQUENCY_MEASURED, KEY_SAMPLING_FREQUENCY, COL_CORE_TICK, \
    KEY_DEVICE_NAME, KEY_IS_LEADING_CORE, METADATA_SEPARATOR, \
    COEFFICIENTS_SEPARATOR, KEY_TIMESTAMP, COL_TIME_EFFECTIVE, COL_TIME

__all__ = ['CoreData']

HeaderType = List[List[AnyStr]]


class CoreData:
    def __init__(self, raw_data_path: str = None,
                 processed_data_path: str = None):
        """
        :param raw_data_path: str = path of the raw data csv
        :param processed_data_path: str = path of the raw data csv
        """

        self._raw = pd.DataFrame()
        self._processed = pd.DataFrame()
        self._tags = set()
        self._metadata = {}
        self._coefficients = {}

        if raw_data_path is not None:
            try:
                self._session_id = get_session_id(raw_data_path)
            except AttributeError:
                self._session_id = -1

            self._raw = self.__read_data_frame(raw_data_path)
            self.__parse_timestamp()

        if processed_data_path is not None:
            self._processed = self.__read_data_frame(processed_data_path)
            self._processed.set_index(
                COL_CORE_TICK,
                drop=True,
                inplace=True
            )

    @property
    def session_id(self) -> int:
        """
        :return: int = session_id for the corresponding session
        """

        return self._session_id

    @property
    def sampling_frequency(self) -> float:
        """
        :return: float = sampling frequency (declared) of the core data
        """

        return float(self._metadata.get(KEY_SAMPLING_FREQUENCY, -1.0))

    @property
    def sampling_frequency_measured(self) -> float:
        """
        :return: float = sampling frequency (measured) of the core data
        """

        return float(self._metadata.get(KEY_SAMPLING_FREQUENCY_MEASURED, -1.0))

    @property
    def core_name(self) -> str:
        """
        :return: str = name of the core
        """

        return self._metadata.get(KEY_DEVICE_NAME, '')

    def is_leading_core(self) -> bool:
        """
        :return: bool = a boolean determining if this core is the leading one
        """

        return as_bool(self._metadata.get(KEY_IS_LEADING_CORE, False))

    @property
    def tags(self) -> Set:
        return self._tags

    @property
    def metadata(self) -> Dict[str, str]:
        """
        :return: dict = the metadata contained in the header of the csv files
        """

        return self._metadata

    @property
    def coefficients(self) -> Dict[str, str]:
        """
        :return: dict = if present the dictionary of the remapping coefficients
                        for the analog inputs
        """

        return self._coefficients

    @property
    def raw(self) -> pd.DataFrame:
        """
        :return: pandas.DataFrame = the raw data for the core
        """

        return self._raw

    @property
    def processed(self) -> pd.DataFrame:
        """
        :return: pandas.DataFrame = the processed data for the core
        """

        return self._processed

    def update_metadata(self, metadata: dict):
        """
        Updates the metadata for the CoreData
        :param metadata: metadata to insert
        """
        self._metadata.update(metadata)

    def add_tag(self, tag: str):
        self._tags.add(tag)

    def remove_tag(self, tag: str):
        self._tags.remove(tag)

    def clear_tags(self):
        self._tags.clear()

    def filter(self, filters: FiltersType):
        """
        Filters the raw data passing them through filters
        :param filters:
            can be:
                - dict with {column_name: Filter}
                    e.g. {'Ax': ButterworthFilter}
                - dict with {column_name: [Filter, Filter, ..]}
                    e.g. {'Ax': [
                                    ButterworthFilter(...),
                                    RecursiveAverageFilter(...)
                                ]}
                    where the filters are applied in sequence
        """

        self._processed = self._raw.copy()

        for column, flt in filters.items():
            if isinstance(flt, list):
                for flt_elem in flt:
                    self.__filter(flt_elem, column=column)
            else:
                self.__filter(flt, column=column)

    def offset(self, offset: Dict[str, float]) -> None:
        """
        Offsets the data using the values in offsets
        :param offset: dict = dict composed as: {column_name : offset}
        """

        for column, off in offset.items():
            if column in self._processed.columns:
                self._processed[column] = self._processed[column] - off

    def subsample(self, target_sampling_frequency=None,
                  decimation_factor: int = None):
        if target_sampling_frequency is None and decimation_factor is None:
            raise ValueError(
                "Must specify one of \'target_sampling_frequency\' "
                "or \'decimation_factor\'")

        if target_sampling_frequency is not None:
            if decimation_factor is None:
                decimation_factor = int(
                    self.sampling_frequency / target_sampling_frequency)
            else:
                raise ValueError(
                    "Must specify just one of \'target_sampling_frequency\' "
                    "or \'decimation_factor\'")

        raw_range = range(0, len(self._raw), decimation_factor)
        processed_range = range(0, len(self._processed), decimation_factor)

        self._raw = self._raw.iloc[raw_range, :]
        self._processed = self._processed.iloc[processed_range, :]

        self._metadata[KEY_SAMPLING_FREQUENCY] = \
            self.sampling_frequency / decimation_factor

        self._metadata[KEY_SAMPLING_FREQUENCY_MEASURED] = \
            self.sampling_frequency_measured / decimation_factor

    def __read_data_frame(self, core_data_path) -> pd.DataFrame:
        header, data_frame = self.__read_csv(core_data_path)

        self.__parse_header(header)
        self.__parse_data(data_frame)
        self.__parse_time(data_frame)

        return data_frame

    @staticmethod
    def __read_csv(core_data_path: str) -> Tuple[HeaderType, pd.DataFrame]:
        header, n_header_rows = read_header(core_data_path)

        raw_data = pd.read_csv(
            filepath_or_buffer=core_data_path,
            skiprows=n_header_rows
        )

        return header, raw_data

    @staticmethod
    def __to_camel_case(text: str):
        if len(text) > 1:
            return text[0].lower() + text[1:]
        else:
            return text

    def __parse_header(self, header: HeaderType):
        self.__parse_metadata(header)
        self.__parse_coefficients(header)

    def __parse_metadata(self, header: HeaderType):
        metadata = [h[0].split(METADATA_SEPARATOR) for h in header if
                    len(h) == 1]
        metadata = [m for m in metadata if len(m) == 2]

        metadata = {
            self.__to_camel_case(m[0].strip()): m[1].strip()
            for m in metadata
        }

        self.update_metadata(metadata)

    def __parse_coefficients(self, header: HeaderType):
        coefficients = [h for h in header if len(h) > 1]

        for line in coefficients:
            row_coefficients = {}
            sensor = line[0]

            for coefficient in line[1:]:
                k, v = coefficient.split(COEFFICIENTS_SEPARATOR)
                row_coefficients.update({k: cast_to_float(v)})

            self._coefficients.update({sensor: row_coefficients})

    def __parse_data(self, data_frame: pd.DataFrame):
        data_frame = data_frame.reset_index(drop=True)

        self.__parse_float_columns(data_frame)

    @staticmethod
    def __parse_float_columns(data_frame: pd.DataFrame):
        for c in data_frame.columns:
            try:
                data_frame[c] = data_frame[c].astype(float)
            except ValueError:
                pass

    def __parse_timestamp(self):
        if KEY_TIMESTAMP not in self._raw.columns:
            return

        n = min(100, len(self._raw))

        timestamp = pd.to_datetime(self._raw[KEY_TIMESTAMP])
        t = (timestamp.values - timestamp.values[0]).astype(float) / 1.0E9

        dt = (t[n:] - t[:-n]) / n
        sampling_frequency = 1.0 / dt.mean()

        self.__check_sampling_frequency(sampling_frequency)
        self.__check_gaps(t)

        self._raw[COL_TIME_EFFECTIVE] = t
        self._raw[KEY_TIMESTAMP] = timestamp.dt.time

        self._metadata.update({
            KEY_SAMPLING_FREQUENCY_MEASURED: sampling_frequency,
        })

    def __check_sampling_frequency(self, sampling_frequency):
        if abs(sampling_frequency / self.sampling_frequency - 1.0) > 0.05:
            logging.warning(
                f'\tsession {self._session_id} | '
                f'Sampling frequency discrepancy: '
                f'measured = {sampling_frequency} '
                f'declared = {self.sampling_frequency}')

    def __check_gaps(self, t):
        dt = t[1:] - t[:-1]
        gaps = dt[(dt / self.sampling_frequency) > 3]

        if len(gaps) > 0:
            logging.warning(
                f'\tsession {self._session_id} | '
                f'Gaps found '
                f'{gaps}')

    def __parse_time(self, data_frame: pd.DataFrame):
        data_frame.drop(columns=['T'], inplace=True, errors="ignore")

        data_frame[COL_TIME] = np.array([
            n / self.sampling_frequency_measured
            for n in range(len(data_frame))
        ])

    def __filter(self, flt: Filter, column: str):
        flt.set_sampling_frequency(self.sampling_frequency)
        self._processed[column] = flt.filter(self._processed[column])



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import core_data
from . import fitfile
from . import session
from . import sessions

from .core_data import *  # NOQA
from .fitfile import *  # NOQA
from .session import *  # NOQA
from .sessions import *  # NOQA

__all__ = []
__all__ += core_data.__all__
__all__ += session.__all__
__all__ += sessions.__all__
__all__ += fitfile.__all__



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from sensorialytics import (sensors, signals, preprocessing, readers,
                            mathematics, helpers)

__all__ = []
__all__ += readers.__all__
__all__ += mathematics.__all__
__all__ += signals.__all__
__all__ += sensors.__all__
__all__ += helpers.__all__
__all__ += preprocessing.__all__



#  constants.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import numpy as np

__all__ = ['PI', 'g', 'GYRO_CONV']

PI: float = np.pi
g: float = 9.81
GYRO_CONV: float = np.pi / 180.0



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import detectors
from . import filters
from . import image

from .detectors import *  # NOQA
from .filters import *  # NOQA
from .image import *  # NOQA

__all__ = []
__all__ += image.__all__
__all__ += detectors.__all__
__all__ += filters.__all__



#  manipulations.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import math
import re
import warnings
from collections import Counter

import numpy
import pandas
from scipy import stats

from sensorialytics.helpers.tools import numeric_types

__all__ = [
    'take_first', 'take_last', 'bin_variable', 'detect_outliers',
    'multiple_replace', 'find_regex', 'fill_empty_with_distr',
    'fill_str', 'norm_cov_matrix'
]


def take_last(s: pandas.Series):
    return s.iloc[-1]


def take_first(s: pandas.Series):
    return s.iloc[0]


def bin_variable(s, min_value, max_value, width, how,
                 default=None, lower=None, upper=None, errors=None):
    """
    :param s: pandas Series or value
    :param min_value: minimum bining value
    :param max_value: maximum binning value
    :param width: bin width
    :param how: function to use for binning. Use numpy.ceil, numpy.floor
    :param default: default value to replace nans
    :param lower: value to assign to values smaller than min_value
    :param upper: value to assign to values larger  than max_value
    :param errors: error values in the Series to be treated a if they where nans
    :return: binned series s_binned
    """
    if isinstance(s, pandas.Series):
        f = __bin_series
    elif isinstance(s, numeric_types):
        f = __bin_value
    else:
        raise TypeError('Unsupported input type {} for s'.format(type(s)))

    result = f(
        s=s,
        min_value=min_value,
        max_value=max_value,
        width=width,
        how=how,
        default=default,
        lower=lower,
        upper=upper,
        errors=errors
    )
    return result


def __bin_series(s: pandas.Series, min_value, max_value, width, how,
                 default, lower, upper, errors):
    s_binned = s.copy()
    nan_values = numpy.array(s_binned.isna()) | numpy.array(
        (s_binned == errors))

    conditions = numpy.array([True for _ in range(len(s_binned))])

    if lower is not None:
        s_binned.loc[s_binned < min_value] = lower
        conditions = conditions & (s_binned != lower) & (min_value <= s_binned)

    if upper is not None:
        s_binned.loc[s_binned > max_value] = upper
        conditions = conditions & (s_binned != upper) & (s_binned <= max_value)

    s_binned.loc[conditions] = list(
        numpy.clip(
            a=how(s_binned.loc[conditions] / width) * width,
            a_min=min_value,
            a_max=max_value)
    )
    s_binned = numpy.round(s_binned, 8)
    if default is not None:
        s_binned.loc[nan_values] = default

    return s_binned


def __bin_value(s: numeric_types, min_value, max_value, width, how,
                default, lower, upper, errors):
    assigned = False
    result = None

    if (math.isnan(s) or s == errors) and default is not None:
        result = default

    else:
        if lower is not None and s < min_value:
            result = lower
            assigned = True
        if upper is not None and s > max_value:
            result = upper
            assigned = True

        if not assigned:
            result = numpy.clip(
                a=how(s / width) * width,
                a_min=min_value,
                a_max=max_value
            )

    return result


def detect_outliers(df, n, features):
    """
    Takes a dataframe df of features and returns a list of the indices
    corresponding to the observations containing more than n outliers according
    to the Tukey method.
    """
    outlier_indices = []

    # iterate over features(columns)
    for col in features:
        # 1st quartile (25%)
        q1 = numpy.percentile(df[col].dropna(), 25)
        # 3rd quartile (75%)
        q3 = numpy.percentile(df[col].dropna(), 75)
        # Interquartile range (iqr)
        iqr = q3 - q1

        # outlier step
        outlier_step = 1.5 * iqr

        # Determine a list of indices of outliers for feature col
        outlier_list_col = df[
            (df[col] < q1 - outlier_step) | (
                    df[col] > q3 + outlier_step)].index

        # append the outlier indices for col to the list of outlier indices
        outlier_indices.extend(outlier_list_col)

    # select observations containing more than 2 outliers
    outlier_indices = Counter(outlier_indices)
    multiple_outliers = list(k for k, v in outlier_indices.items() if v > n)

    return multiple_outliers


def multiple_replace(s: pandas.Series, replacing_dict: dict, default=None):
    """
    replaces multiple values at once
    :param s: pandas Series to act on
    :param replacing_dict: dictionary with replacements. Has to be in the form:
     {
        '<replacing_value_1>' : [<list_of_values_to_replace],
        '<replacing_value_2>' : [<list_of_values_to_replace] ...
     }
    :param default: default value. used to fill values not in replacing_value_n
    :return:
    """
    for key, value in replacing_dict.items():
        s = s.replace(value, key)
    if default is not None:
        s = s.map(
            dict(zip(replacing_dict.keys(), replacing_dict.keys()))).fillna(
            default)
    return s


def find_regex(regex, value):
    if re.search(regex, value) is not None:
        return True
    else:
        return False


def fill_empty_with_distr(s: pandas.Series, values_to_fill: list,
                          values_to_avoid: list = None):
    """
    :param s: pandas Series to be filled
    :param values_to_fill: value to fill
    :param values_to_avoid: values not valid for filling
    :return:
    """

    compl_values_to_avoid = [numpy.nan]
    if values_to_avoid is not None:
        compl_values_to_avoid += values_to_avoid

    filling_index = s.index[s.isin(values_to_fill)]

    distrib_data = s.value_counts()
    distrib_data = distrib_data[
        ~distrib_data.index.isin(values_to_fill + compl_values_to_avoid)]
    distrib_data_dict = {k: v for k, v in enumerate(distrib_data.index)}

    xk = list(distrib_data_dict.keys())
    pk = distrib_data.values / distrib_data.values.sum()

    distrib = stats.rv_discrete(values=(xk, pk))

    filled_s = s.copy()
    filled_s[filling_index] = [distrib_data_dict[distrib.rvs()] for _ in
                               filling_index]

    return filled_s


def fill_str(s: pandas.Series, length, fill_value: str = '0', fillna=False):
    if s.isna().any() and not fillna:
        warnings.warn('{} NaNs found in input series. Mapping skipped'.format(
            sum(s.isna())), Warning)

    return s.astype(float).apply(
        lambda x: str(int(x)).rjust(length, fill_value) if not math.isnan(
            x) else ''.rjust(length, fill_value) if fillna else x)


def norm_cov_matrix(df):
    """
    :param df: pandas dataframe to use
    :return: normalized covariance matrix of a pandas dataframe
    """
    var = numpy.sqrt(df.var())
    return (df.loc[:, var.index] / var).cov()



#  timer.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import datetime


class ExecutionTimer:
    __start = datetime.datetime.now()

    @staticmethod
    def reset():
        ExecutionTimer.__start = datetime.datetime.now()

    @staticmethod
    def get_elapsed_time() -> datetime.timedelta:
        return datetime.datetime.now() - ExecutionTimer.__start



#  clustering.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from itertools import groupby

import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn
from sklearn.cluster import SpectralClustering, DBSCAN

from .manipulations import norm_cov_matrix

__all__ = ['cluster_variables']


def cluster_variables(df: pandas.DataFrame, n_clusters: int, sigma: float,
                      clustering_variables: set = None, target: set = None,
                      return_results: bool = False,
                      plot_results: bool = False,
                      **kwargs):
    """
    :param df: pandas dataframe
    :param n_clusters: number of cluster to create
    :param sigma: decay constant for affinity matrix
    :param clustering_variables: variables to use in clustering
    :param target: optional target variable to spot correlations
    :param return_results: return full clusters description
    :param plot_results: plot the covariance and affinity matrices
    :param kwargs: kwargs for seaborn heatmap
    :return: full clusters description
    """
    # Affinity -----------------------------------------------------------------
    clustering_variables, matrix_variables, target = __get_variables(
        clustering_variables=clustering_variables,
        df=df,
        target=target
    )

    clustering_matrix_full, clustering_matrix, cov_matrix = __get_matrices(
        clustering_variables=clustering_variables,
        df=df,
        matrix_variables=matrix_variables,
        sigma=sigma,
        n_clusters=n_clusters
    )

    # Clusters -----------------------------------------------------------------
    clusters, clusters_dim, ordered_variables = __get_clusters(
        clustering_matrix=clustering_matrix,
        clustering_variables=clustering_variables,
        n_clusters=n_clusters,
        sigma=sigma,
        target=target)

    sorted_cov_matrix = cov_matrix.loc[ordered_variables, ordered_variables]
    sorted_clustering_matrix_full = clustering_matrix_full.loc[
        ordered_variables, ordered_variables]

    # Plots --------------------------------------------------------------------
    if plot_results:
        __plot_results(
            sorted_cov_matrix=sorted_cov_matrix,
            sorted_clustering_matrix_full=sorted_clustering_matrix_full,
            clusters_dim=clusters_dim,
            target=target,
            **kwargs)

    if return_results:
        return {
            'clusters': clusters,
            'cov_matrix': cov_matrix,
            'clustering_matrix': clustering_matrix,
            'sorted_cov_matrix': sorted_cov_matrix,
            'sorted_clustering_matrix_full': sorted_clustering_matrix_full
        }


def __get_matrices(clustering_variables, df, matrix_variables, sigma,
                   n_clusters):
    cov_matrix = numpy.abs(norm_cov_matrix(df[matrix_variables]))
    dist_matrix = numpy.maximum(1. - cov_matrix, 0.0)

    clustering_matrix_full = __get_clustering_matrix_full(
        n_clusters=n_clusters,
        dist_matrix=dist_matrix,
        sigma=sigma
    )

    clustering_matrix = \
        clustering_matrix_full.loc[clustering_variables, clustering_variables]

    return clustering_matrix_full, clustering_matrix, cov_matrix


def __get_clustering_matrix_full(n_clusters, dist_matrix, sigma):
    if n_clusters is None:
        return dist_matrix
    else:
        return numpy.exp(- dist_matrix ** 2 / (2. * sigma ** 2))


def __get_variables(clustering_variables, df, target):
    if target is None:
        target = set()

    if clustering_variables is None:
        matrix_variables = set(df.columns)
    else:
        matrix_variables = clustering_variables.copy()

    matrix_variables.update(target)
    clustering_variables = matrix_variables.copy()
    clustering_variables.difference_update(target)

    return clustering_variables, matrix_variables, target


def __get_clusters(clustering_matrix, clustering_variables, n_clusters, sigma,
                   target):
    if n_clusters is None:
        cluster = DBSCAN(metric='precomputed', eps=sigma, min_samples=1)
    else:
        cluster = SpectralClustering(affinity='precomputed',
                                     n_clusters=n_clusters)

    clusters_tags = cluster.fit_predict(X=clustering_matrix)

    clusters = {}
    for k, v in sorted(zip(clustering_variables, clusters_tags)):
        clusters.setdefault(v, []).append(k)

    clusters_dim = [
        len(list(group))
        for _, group in groupby(sorted(clusters_tags))
    ]

    ordered_variables = [
                            x for _, x in
                            sorted(zip(clusters_tags, clustering_variables))
                        ] + list(target)

    return clusters, clusters_dim, ordered_variables


def __plot_results(sorted_cov_matrix, sorted_clustering_matrix_full,
                   clusters_dim,
                   target, **kwargs):
    height = max(10, len(sorted_cov_matrix.columns) // 5)
    plt.rcParams['figure.figsize'] = (18, height)

    titles = ['sorted_cov_matrix', 'sorted_clustering_matrix']
    matrices = [sorted_cov_matrix, sorted_clustering_matrix_full]

    for title, matrix in zip(titles, matrices):
        g = seaborn.heatmap(
            data=matrix,
            cmap='viridis',
            linecolor='black',
            linewidths=0.1,
            **kwargs)

        g.axes.hlines(numpy.cumsum(clusters_dim)[:-1], *g.axes.get_xlim(),
                      colors='white', )
        g.axes.vlines(numpy.cumsum(clusters_dim)[:-1], *g.axes.get_xlim(),
                      colors='white')

        if len(target) > 0:
            g.axes.hlines(numpy.cumsum(clusters_dim)[-1],
                          *g.axes.get_xlim(),
                          colors='red', )
            g.axes.vlines(numpy.cumsum(clusters_dim)[-1],
                          *g.axes.get_xlim(),
                          colors='red')

        g.set_title(title)
        plt.show()



#  tools.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import json
import logging
import os

import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn as sns

from sensorialytics.helpers.exceptions import MaximumDepthError

__all__ = ['float_types', 'int_types', 'bool_types', 'collection_types',
           'numeric_types', 'pseudo_numeric_types', 'na_values', 'as_bool',
           'get_json_dumpable', 'jprint', 'find_key', 'get_keys', 'describe']

float_types = (float, numpy.float, numpy.float16, numpy.float32, numpy.float64)
int_types = (int, numpy.int, numpy.int8, numpy.int16, numpy.int32, numpy.int64)
bool_types = (bool, numpy.bool_)
collection_types = (dict, list, set, numpy.ndarray)

numeric_types = float_types + int_types
pseudo_numeric_types = float_types + int_types + bool_types

na_values = [
    '', '#N/A', '#N/A N/A', '#NA', '-1.#IND', '-1.#QNAN', '-NaN', '-nan',
    '1.#IND', '1.#QNAN', 'N/A', 'NULL', 'NaN', 'n/a', 'nan', 'null'
]


def list_tree(d):
    if os.path.isfile(d):
        return d.rsplit('/')[-1]
    if os.path.isdir(d):
        return {d: [list_tree(d + '/' + dd) for dd in os.listdir(d)]}


def list_files(tree):
    if isinstance(tree, dict):
        for k, v in tree.items():
            if isinstance(v, list):
                return [list_files(t) for t in v]
    if isinstance(tree, str):
        return [tree]


def as_bool(x) -> bool:
    if isinstance(x, bool):
        return x

    if x in ['true', 'True']:
        return True

    if x in ['false', 'False']:
        return False

    return bool(float(x))


def get_json_dumpable(data, depth: int = 0, max_depth: int = None):
    if isinstance(data, dict):
        result = {
            __process_value(k, depth=depth, max_depth=max_depth):
                __process_value(v, depth=depth, max_depth=max_depth)
            for k, v in data.items()
        }
    elif isinstance(data, (list, set, numpy.ndarray)):
        result = [
            __process_value(x, depth=depth, max_depth=max_depth) for x in data
        ]
    elif isinstance(data, pseudo_numeric_types):
        if isinstance(data, int_types):
            result = int(data)
        elif isinstance(data, float_types):
            result = float(data)
        else:
            result = bool(data)
    else:
        raise TypeError('Unsupported type {}'.format(type(data)))

    return result


def __process_value(v, depth: int, max_depth: int):
    if depth == max_depth:
        raise MaximumDepthError('Maximum depth {} reached'.format(max_depth))
    if isinstance(v, (str, type, pandas.datetime, Exception)):
        v_new = str(v)
    elif isinstance(v, bool_types):
        v_new = bool(v)
    elif isinstance(v, float_types):
        v_new = float(v)
    elif isinstance(v, int_types):
        v_new = int(v)
    elif isinstance(v, collection_types):
        v_new = get_json_dumpable(
            data=v.copy(),
            depth=depth + 1,
            max_depth=max_depth
        )
    elif v is None:
        v_new = v
    else:
        raise TypeError('Unsupported type {}'.format(type(v)))
    return v_new


def jprint(j, indent=2, **kwargs):
    print(json.dumps(get_json_dumpable(j, *kwargs), indent=indent))


def find_key(x: (dict, list), key: str, broad=False) -> list:
    """
    :param x: dictionary or list containing dictionaries
    :param key: key to look for
    :param broad: broad match for the value
    :return:
    """
    res = []

    if isinstance(x, dict):
        match = __match_key(d=x, key=key, broad=broad)
        if match[0]:
            found = [
                find_key(x=x[m], key=key, broad=broad)
                for m in match[1]
            ]

            res = [
                list((m, f)) if len(f) > 1
                else m[0] if len(m) == 1 else m
                for m, f in zip(match[1], found)
            ]

        else:
            for k, v in x.items():
                if isinstance(v, (dict, list)):
                    res += [
                        [k] + e if isinstance(e, list)
                        else [k] + [e]
                        for e in find_key(x=v, key=key, broad=broad)]

    elif isinstance(x, list):
        for i, v in enumerate(x):
            res += [
                [[i]] + e if isinstance(e, list)
                else [[i]] + [e]
                for e in find_key(x=v, key=key, broad=broad)
            ]
    return res


def __match_key(d: dict, key: str, broad: bool):
    if broad:
        b_match = [k for k in d.keys() if key in k]
        return any(b_match), b_match
    else:
        return key in d.keys(), [key]


def get_keys(d: dict, key: str, broad: bool = False) -> list:
    keys = find_key(x=d, key=key, broad=broad)
    result = []
    for k in keys:
        elem = ""
        for e in k:
            if isinstance(e, list):
                z = str(e[-1])
            else:
                z = "'{}'".format(e)
            elem += "[{}]".format(z)
        result.append(elem)
    return result


def describe(
        df: pandas.DataFrame, columns=None, excluded_columns=None,
        n_max_levels=100, n_bins=100):
    """
    :param df: pandas.DataFrame to plot
    :param columns: columns to plot. None plots all the columns in df
    :param excluded_columns: columns to exclude
    :param n_max_levels: maximum number of levels to plot. Variables with more
            than n_max_levels will not be plot
    :param n_bins: nunber of bins for float/int variables
    """
    original_figsize = plt.rcParams['figure.figsize']
    plt.rcParams['figure.figsize'] = (18, original_figsize[1])

    if columns is None:
        columns = df.columns
    if excluded_columns is not None:
        columns = [x for x in columns if x not in excluded_columns]

    for col in columns:
        dtype = df[col].dtype
        uniques = df[col].unique()
        n_uniques = len(uniques)

        info = f'[variable: {col}] [dtype: {dtype}] [n_uniques: {n_uniques}]'
        logging.info('{:_<125} '.format(info))

        if dtype in float_types:
            df[col].hist(bins=n_bins)
            plt.tick_params('x', rotation=45)
        elif dtype in int_types:
            if n_uniques <= n_max_levels:
                sns.countplot(df[col])
                plt.tick_params('x', rotation=45)
            else:
                df[col].hist(bins=n_bins)
                plt.tick_params('x', rotation=45)
        elif dtype == 'O' and n_uniques < n_max_levels:
            sns.countplot(df[col])
            plt.tick_params('x', rotation=45)

        plt.show()

    plt.rcParams['figure.figsize'] = original_figsize



#  scalers.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import numpy
import pandas

from sensorialytics.helpers.exceptions import FitError

__all__ = ['Scaler', 'MultiDimStandardScaler']


class Scaler:
    def __init__(self, scaler_class, fit_variables):
        self.__scaler_class = scaler_class
        self.__fit_variables = fit_variables
        self.__scalers = {}
        self.__fitted = False

    def reset(self):
        self.__scalers = {}
        self.__fitted = False

    def fit(self, df: pandas.DataFrame):
        self.reset()

        for variable in self.__fit_variables:
            scaler = self.__scaler_class()
            scaler.fit(df[[variable]])
            self.__scalers.update({variable: scaler})

        self.__fitted = True

    def transform(self, data, transform_variables: list = None,
                  array_indices: list = None):
        if transform_variables is None:
            transform_variables = self.__fit_variables
        if self.__fitted:
            if isinstance(data, dict):
                transformed = self.__transform_dict(data, transform_variables)

            elif isinstance(data, pandas.DataFrame):
                transformed = self.__transform_data_frame(data,
                                                          transform_variables)

            elif isinstance(data, pandas.Series):
                transformed = self.__transform_series(data, transform_variables)

            elif isinstance(data, numpy.ndarray):
                transformed = self.__transform_array(array_indices, data,
                                                     transform_variables)
            else:
                raise TypeError(f'Type {type(data)} is invalid for scaling '
                                f'transformation')

        else:
            raise FitError('Scaler not fitted')

        return transformed

    def __transform_array(self, array_indices, data, transform_variables):
        if len(data.shape) != 2:
            raise ValueError(f'2D array expected. '
                             f'Got {data.shape}D array instead')

        if len(array_indices) == len(transform_variables):
            transformed = numpy.array([]).reshape(data.shape[0], 0)
            for i, index in enumerate(range(data.shape[1])):
                if index in array_indices:
                    variable = transform_variables[i]
                    scaler = self.__scalers[variable]
                    transformed = numpy.concatenate(
                        arrays=[
                            transformed,
                            scaler.transform(data[:, [index]])
                        ],
                        axis=1
                    )
                else:
                    transformed = numpy.concatenate(
                        arrays=[transformed, data[:, [i]]],
                        axis=1)
        else:
            raise ValueError(
                'Wrong number of array_indices! Expected {}, got {}'.format(
                    len(transform_variables), len(array_indices)))

        return transformed

    def __transform_series(self, data, transform_variables):
        original_variables = list(data.index)
        other_variables = list(data.index.difference(transform_variables))
        transformed = pandas.Series(name=data.name)
        for variable in transform_variables:
            transformed = transformed.append(pandas.Series(
                self.__scalers[variable].transform([[data[variable]]])[0],
                index=[variable]))
        transformed = data[other_variables].append(transformed)[
            original_variables]

        return transformed

    def __transform_data_frame(self, data, transform_variables):
        original_variables = list(data.columns)
        other_variables = list(data.columns.difference(transform_variables))
        transformed = []
        for variable in transform_variables:
            scld = self.__scalers[variable].transform(data[[variable]])
            transformed.append(
                pandas.DataFrame(
                    data=scld,
                    index=data.index,
                    columns=[variable]
                )
            )
        transformed.append(data[other_variables])
        transformed = pandas.concat(transformed, axis=1)[original_variables]

        return transformed

    def __transform_dict(self, data, transform_variables):
        transformed = data.copy()
        for variable in transform_variables:
            scld = self.__scalers[variable].transform([[data[variable]]])[0, 0]
            transformed.update({
                variable: scld
            })

        return transformed

    def fit_transform(self, df: pandas.DataFrame):
        self.fit(df)

        return self.transform(df)


class MultiDimStandardScaler:
    def __init__(self):
        self._mean = None
        self._var = None
        self._std = None
        self.__fitted = None

    def fit(self, df):
        if isinstance(df, pandas.DataFrame):
            data = df.values
        else:
            data = df.copy()

        self._mean = numpy.mean(data, axis=0)
        self._var = numpy.var(data, axis=0)
        self._std = numpy.sqrt(self._var)
        self.__fitted = True

    def transform(self, df):
        if self.__fitted:
            sample_len = len(df)
            scaled = df - numpy.array([self._mean for _ in range(sample_len)])
            scaled = scaled / numpy.array(
                [self._std for _ in range(sample_len)])
        else:
            raise FitError('Scaler not fitted.')

        return scaled

    def fit_transform(self, df):
        self.fit(df)

        return self.transform(df)



#  tools.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import numpy as np

from sensorialytics.mathematics.point import Point
from sensorialytics.mathematics.polynomial import Polynomial

__all__ = ['cast_to_float', 'is_iterable', 'derive', 'simpson_derivative',
           'sign', 'unwrap', 'get_line', 'get_parallel_line',
           'get_lines_intersection']


def cast_to_float(x):
    if is_iterable(x) and not isinstance(x, str):
        return np.array([cast_to_float(v) for v in x])
    else:
        try:
            return float(x)
        except ValueError:
            return np.nan


def is_iterable(obj):
    try:
        iter(obj)
    except TypeError:
        return False
    else:
        return True


def derive(x: np.ndarray, dx: float = 1.0, shift: int = 1):
    return np.append([0], x[shift:] - x[:-shift]) / (shift * dx)


def simpson_derivative(x, dt):
    derivative = np.zeros(x.shape)

    derivative[2:-2] = (x[:-4] - 8 * x[1:-3] + 8 * x[3:-1] - x[4:]) / (12 * dt)

    return derivative


def sign(x):
    return int(x > 0)


def unwrap(x: np.ndarray, mod: float):
    """
    Inverse of np.mod
    :param x: array to unwrap
    :param mod:modulus used to wrap
    :return: unwrapped np.ndarray
    """
    half_mod = mod / 2.0
    unwrapped = np.array([])
    delta = 0.0
    old_x = x[0]

    for v in x:
        if old_x - v > half_mod:
            delta += mod

        unwrapped = np.append(unwrapped, [v + delta])
        old_x = v

    return unwrapped


def get_line(point1: Point, point2: Point) -> Polynomial:
    if point2[0] == point1[0]:
        raise ZeroDivisionError('Can\'t determine the equation if p1_x = p2_x')

    m = (point2[1] - point1[1]) / (point2[0] - point1[0])
    q = point1[1] - m * point1[0]

    return Polynomial(q, m)


def get_parallel_line(line: Polynomial, point: Point) -> Polynomial:
    if line.degree != 1:
        raise RuntimeError('Provided polynomial is not a line')

    m = line[1]
    q = point[1] - m * point[0]

    return Polynomial(q, m)


def get_lines_intersection(line1: Polynomial, line2: Polynomial) -> Point:
    if line1.degree != 1 or line2.degree != 1:
        raise RuntimeError('Polynomials have degree != 1')

    if line1[1] == line2[1]:
        raise RuntimeError('Lines are parallel')

    x = (line2[0] - line1[0]) / (line1[1] - line2[1])
    y = line1[1] * x + line1[0]

    return Point(x, y)



#  point.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved
import numpy as np

from sensorialytics.helpers.tools import numeric_types

__all__ = ['Point']


class Point:
    def __init__(self, *coordinates):
        self._dimension = 0
        self._coordinates = np.array([])
        self.__unpack_coordinates(*coordinates)

    def __getitem__(self, item: int) -> float:
        return self._coordinates[item]

    def __str__(self):
        to_str = ''

        if len(self._coordinates) != 0:
            to_str = ', '.join([str(np.round(x, 5)) for x in self._coordinates])

        return f'({to_str})'

    def __mul__(self, other):
        if isinstance(other, Point):
            if other.dimension != self._dimension:
                raise ValueError(f'Points have different dimension! '
                                 f'{self._dimension} != {other.dimension}')
            return Point([
                self._coordinates[i] * other._coordinates[i]
                for i in range(self._dimension)
            ])
        elif isinstance(other, numeric_types):
            return Point([
                self._coordinates[i] * other
                for i in range(self._dimension)
            ])

    def __rmul__(self, other):
        return self.__mul__(other)

    @property
    def dimension(self):
        return self._dimension

    @property
    def coordinates(self) -> np.ndarray:
        return self._coordinates

    def __unpack_coordinates(self, *coordinates):
        for c in coordinates:
            if isinstance(c, (tuple, list, np.ndarray)):
                self._coordinates = np.append(self._coordinates, c)
                self._dimension += len(c)
            else:
                self._coordinates = np.append(self._coordinates, [c])
                self._dimension += 1



#  functions.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from typing import Union

import numpy as np

__all__ = ['Function', 'Sigmoid', 'SignSigmoid', 'Step', 'Characteristic',
           'Symmetrised']

FunctionInput = Union[np.ndarray, float, int]

EXP_ARG_OVERFLOW = np.log(np.finfo(np.float).max)


class Function:
    def __init__(self):
        pass

    def __call__(self, x: FunctionInput):
        return x


class Sigmoid(Function):
    def __init__(self, sigma: float, x0: float = 0.0):
        super().__init__()
        self._sigma = sigma
        self._x0 = x0

    def __call__(self, x: FunctionInput):
        k = np.minimum(x, EXP_ARG_OVERFLOW)
        e = np.exp(k)

        return e / (1 + e)


class SignSigmoid(Sigmoid):
    def __init__(self, sigma: float, x0: float = 0.0):
        super().__init__(sigma, x0)

    def __call__(self, x: FunctionInput):
        return 2.0 * super()(x) - 1.0


class Step(Function):
    def __init__(self, threshold: float = 0.0):
        super().__init__()

        self._threshold = threshold

    def __call__(self, x: FunctionInput):
        if isinstance(x, np.ndarray):
            return (x >= self._threshold).astype(float)
        else:
            return float(x >= self._threshold)


class Characteristic(Function):
    def __init__(self, lower, upper):
        super().__init__()

        if lower < upper:
            self._lower = Step(lower)
            self._upper = Step(upper)
        else:
            raise ValueError('must be lower < upper')

    def __call__(self, x: FunctionInput):
        return self._lower(x) - self._upper(x)


class Symmetrised(Function):
    def __init__(self, function: Function):
        super().__init__()

        self._base_function = function

    def __call__(self, x: FunctionInput):
        pos = self._base_function(+x) * (x >= 0)
        neg = self._base_function(-x) * (x < 0)

        return pos + neg



#  setup.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from setuptools import setup, find_packages

version = '1.0.4'

setup(
    name='sensorialytics',
    packages=find_packages(),
    version=version,
    license='MIT',
    description='Sensoria python library',
    author='Stefano Rossotti',
    author_email='stefano@sensoriainc.com',
    keywords=['sensoria', 'analytics'],
    install_requires=[
        "numpy>=1.16.5",
        "requests",
        "requests-oauthlib",
        "fitdecode",
        "sklearn",
        "pandas>=1.1.3",
        "matplotlib>=3.3.2",
        "scipy>=1.5.2",
        "opencv-python",
        "seaborn",
        "xgboost",
        "keras",
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import dates
from . import sensoria_io_client
from . import tools

from .dates import *  # NOQA
from .sensoria_io_client import *  # NOQA
from .tools import *  # NOQA

__all__ = []
__all__ += dates.__all__
__all__ += tools.__all__
__all__ += sensoria_io_client.__all__



__all__ = ['Sensor']


#  sensor.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

class Sensor:
    def __init__(self) -> None:
        super().__init__()



#  pressure.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from typing import Union

import numpy as np
import pandas as pd

from .sensor import Sensor

__all__ = ['RemapInput', 'Pressure', 'FSR', 'MatTextile']

RemapInput = Union[np.ndarray, float, int, pd.Series]
RemapOutput = Union[np.ndarray, float, int, pd.Series]


class Pressure(Sensor):
    def __init__(self, remap_parameters: dict = None) -> None:
        super().__init__()

        self._offset = 0.0

        if remap_parameters is not None:
            self._remap_parameters = remap_parameters
        else:
            self._remap_parameters = {}

    def set_remap_parameters(self, remap_parameters: dict):
        self._remap_parameters = remap_parameters

        for name, value in self._remap_parameters.items():
            self.__setattr__(name, value)

    def remap(self, x: RemapInput) -> RemapOutput:
        return x

    def offset(self, o: float):
        self._offset = o

    def get(self, k: str) -> object:
        return self._remap_parameters.get(k)


class FSR(Pressure):
    def __init__(self, remap_parameters: dict = None):
        super().__init__(remap_parameters=remap_parameters)

    def remap(self, x: RemapInput) -> RemapOutput:
        a = self._remap_parameters.get('A')
        b = self._remap_parameters.get('B')
        c = self._remap_parameters.get('C')

        return a * (np.exp(b * (x - c)) - 1.0) - self._offset


class MatTextile(Pressure):
    def __init__(self, remap_parameters: dict = None):
        super().__init__(remap_parameters=remap_parameters)

    def remap(self, x: RemapInput) -> RemapOutput:
        a = self._remap_parameters.get('A')
        b = self._remap_parameters.get('B')
        c = self._remap_parameters.get('C')

        return a * np.power(x, -b) - c - self._offset



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import clustering
from . import encoders
from . import manipulations
from . import scalers
from . import tools

from .clustering import *  # NOQA
from .encoders import *  # NOQA
from .manipulations import *  # NOQA
from .scalers import *  # NOQA
from .tools import *  # NOQA

__all__ = []
__all__ += clustering.__all__
__all__ += encoders.__all__
__all__ += manipulations.__all__
__all__ += scalers.__all__
__all__ += tools.__all__



#  sensoria_io_client.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import getpass
import logging

import requests
from requests_oauthlib import OAuth2

__all__ = ['SensoriaIoClient', 'TokenRetrievingFailed']

N_MAX_AUTH_TRIALS = 5
TOKEN_URL = 'https://auth.sensoriafitness.com/oauth20/token/'
TOKEN_HEADERS = {
    'Host': 'auth.sensoriafitness.com',
    'Content-Type': 'application/x-www-form-urlencoded',
    'Accept': 'application/x-www-form-urlencoded, application/json',
    'Authorization': 'Basic '
                     'NjM1MjE3NDE5NzAwN'
                     'jA5ODc1OjQ4ZWIxY2'
                     'E2OTkyODQ1ODU5M2J'
                     'mYjc1NTEyYWM4ZmMx'
}


class SensoriaIoClient:
    __token = None

    @staticmethod
    def token() -> OAuth2:
        if not SensoriaIoClient.__has_token():
            SensoriaIoClient.__authenticate()

        return SensoriaIoClient.__token

    @staticmethod
    def __has_token() -> bool:
        return SensoriaIoClient.__token is not None

    @staticmethod
    def __authenticate():
        n_trials = 0

        while n_trials < N_MAX_AUTH_TRIALS and \
                not SensoriaIoClient.__has_token():
            n_trials += 1

            username = input('username: ')
            password = getpass.getpass('password: ')

            try:
                SensoriaIoClient.__token = SensoriaIoClient.__get_token(
                    username=username,
                    password=password
                )
            except TokenRetrievingFailed as e:
                logging.error(f'{e}. {N_MAX_AUTH_TRIALS - n_trials} '
                              f'attempts remaining')

    @staticmethod
    def __get_token(username: str, password: str) -> OAuth2:
        response = SensoriaIoClient.__get_token_response(password, username)

        if response.reason == 'OK':
            access_token = response.json()['access_token']

            return OAuth2(token={'access_token': access_token})
        else:
            raise TokenRetrievingFailed(f'Token retrieving failed. '
                                        f'Reason: {response.reason}')

    @staticmethod
    def __get_token_response(password, username) -> requests.Response:
        user, domain = username.split('@')
        body = (f'grant_type=password&username={user}%40{domain}&'
                f'password={password}&scope=sessions.read%20sessions.write%20'
                f'users.read%20users.write%20workspaces.read%20'
                f'workspaces.write%20shoes.read%20shoes.write%20shoes.delete%20'
                f'firmware.read%20settings.read',)[0]

        token_response = requests.post(
            url=TOKEN_URL,
            headers=TOKEN_HEADERS,
            data=body.encode('utf-8')
        )

        return token_response


class TokenRetrievingFailed(Exception):
    def __init__(self, message: str):
        super().__init__(message)



#  helper.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import re
from typing import Dict, List, Union

from sensorialytics.signals.filters import Filter

SESSION_DIR_REGEX = r'\bsession-[0-9]{6}-raw$'
SESSION_DIR_REGEX_BROAD = r'\bsession-[0-9]{6}-raw'
SESSION_NAME_REGEX = [r'\b[0-9]{6}-[0-9]x[0-9]{4}.csv$',
                      r'\b[0-9]{6}-[0-9]{5}-[0-9]x[0-9]{4}.csv$']
SESSION_ID_REGEX = r'[0-9]{6}'

KEY_CSV_LOG_TYPE = "LogType"
KEY_CSV_DEVICE_NAME = "DeviceName"
METADATA_SEPARATOR = ":"
COEFFICIENTS_SEPARATOR = ":"

LOG_TYPE_PROCESSED = "Processed"
LOG_TYPE_RAW = "Raw"
LOG_TYPE_AGGREGATE = "Aggregate"
LOG_TYPE_EVENTS = "Events"
LOG_TYPE_CORE_DATA = "CoreData"

KEY_SAMPLING_FREQUENCY = "samplingFrequency"
KEY_SAMPLING_FREQUENCY_MEASURED = "samplingFrequencyMeasured"
KEY_DEVICE_NAME = "deviceName"
KEY_IS_LEADING_CORE = "isLeadingCore"
KEY_TIMESTAMP = "Timestamp"
KEY_ON_SETTINGS_SAVED = "onSettingsSaved"

COL_LEADING_CORE_TICK = "LeadingCoreTick"
COL_CORE_TICK = "CoreTick"
COL_EVENT = "Event"
COL_EVENT_NAME = "EventName"
COL_EVENT_PARAMS = "EventParameters"
COL_EVENT_INFO = "EventInfo"
COL_TIME = 't'
COL_TIME_EFFECTIVE = 'tEffective'

FiltersType = Union[Dict[str, Filter], Dict[str, List[Filter]]]


def is_session_name(filename: str) -> bool:
    return any([re.search(r, filename) for r in SESSION_NAME_REGEX])


def get_session_id(session_dir: str) -> int:
    try:
        session_match = re.search(SESSION_DIR_REGEX, session_dir).group(0)
    except AttributeError:
        session_match = re.search(SESSION_DIR_REGEX_BROAD, session_dir).group(0)

    session_id = int(re.search(SESSION_ID_REGEX, session_match).group(0))

    return session_id


def read_header(core_data_path):
    with open(core_data_path, 'r') as f:
        lines = f.readlines()

    header = [
        (i, line) for i, line in enumerate(lines)
        if len(line.split(',')) == 1 and line != 'sep=,\n'
    ]

    n_header_rows = header[-1][0] + 1

    header = [
        line[1].replace(',\n', '').replace('\n', '').split(',')
        for line in header
    ]

    return header, n_header_rows



#  __init__.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

from . import pressure
from . import sensor

from .pressure import *  # NOQA
from .sensor import *  # NOQA

__all__ = []
__all__ += sensor.__all__
__all__ += pressure.__all__



#  sessions.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import logging
import os
import pickle
import re
from os.path import join
from typing import Dict, Set

import requests

from sensorialytics.helpers.sensoria_io_client import SensoriaIoClient
from .helper import SESSION_DIR_REGEX, FiltersType, get_session_id
from .session import Session

__all__ = ['Sessions']

SESSIONS_FILENAME = 'sessions.pickle'
SESSION_URL = 'https://api.sensoriafitness.com/api/1.0/session/'


class Sessions(Dict[int, Session]):
    SESSION_CLASS = Session

    def __init__(self, sessions_dir, read_titles: bool = True,
                 save: bool = True, load: bool = False, update: bool = True):
        """
        :param sessions_dir: directory containing the sessions to handle
        :param read_titles: if true interrogates the cloud to get the tiles
        :param save: saves in session_dir a .pickle file containing self
        :param load: if present loads the .pickle file instead of reading the
                sessions. If a new session is inserted in session_dir it will be
                read
        """
        super().__init__()

        self.__sessions_dir = sessions_dir

        found_ids = set()

        if load:
            found_ids = self.__load_sessions()

        if update:
            self.__read_sessions(found_ids, read_titles, save)

    def __iter__(self):
        for session_id, session in self.items():
            yield session_id, session

    def __str__(self):
        return str(self.ids)

    @property
    def ids(self):
        """
        :return: list of session ids
        """
        return list(self.keys())

    @property
    def titles(self):
        """
        list of titles of the sessions
        :return:
        """
        return [session.title for session in self.values()]

    @property
    def description(self):
        """
        :return: dict with session id and title for each session
        """
        return {session.session_id: session.title for session in self.values()}

    def save(self, tag: str = None):
        """
        Saves the sessions in self.__sessions_dir provided in the constructor
        :param tag: str optional =  tag for the specific saved instance
        """

        if tag is not None:
            filename = f'{SESSIONS_FILENAME}-{tag}'
        else:
            filename = SESSIONS_FILENAME

        with open(join(self.__sessions_dir, filename), 'wb') as f:
            pickle.dump(self, f)

    def merge(self, sessions):
        for session in sessions.values():
            self.insert_session(session)

    def insert_session(self, session: Session):
        if session.session_id not in self.keys():
            if isinstance(session, self.SESSION_CLASS):
                self.update({session.session_id: session})
            else:
                raise ValueError(
                    f"Can't add a session of type {type(session)} "
                    f"to a session set of type {type(self.SESSION_CLASS)}")

    def filter(self, filters: FiltersType):
        """
        Filters the raw data passing them through filters
        :param filters:
            can be:
                - dict with {column_name: Filter}
                    e.g. {'Ax': ButterworthFilter}
                - dict with {column_name: [Filter, Filter, ..]}
                    e.g. {'Ax': [
                                    ButterworthFilter(...),
                                    RecursiveAverageFilter(...)
                                ]}
                    where the filters are applied in sequence
        """

        for session in self.values():
            session.filter(filters=filters)

    def offset(self, offsets: Dict[str, Dict[str, float]]) -> None:
        """
        Offsets the data using the values in offsets
        :param offsets: dict composed as:
            {
                core_id : {
                    column_name : offset
                }
            }
        """

        for session in self.values():
            session.offset(offsets=offsets)

    def subsample(self, target_sampling_frequency=None,
                  decimation_factor: int = None, save=False):
        for session in self.values():
            session.subsample(
                target_sampling_frequency=target_sampling_frequency,
                decimation_factor=decimation_factor
            )

        if save:
            self.save()

    def __read_sessions(self, found_ids: set,
                        read_titles: bool, save: bool):
        for session_name in os.listdir(self.__sessions_dir):
            if re.search(SESSION_DIR_REGEX, session_name) is not None:
                session_dir = join(self.__sessions_dir, session_name)
                session_id = get_session_id(session_dir)

                if session_id not in found_ids:
                    self.__read_session(session_dir, read_titles)

        if save:
            self.save()

        logging.info('Done!')

    def __read_session(self, session_dir, read_titles):
        session = self.SESSION_CLASS(session_dir=session_dir)

        self.insert_session(session)

        if read_titles:
            self.__read_title(session=session)

    @staticmethod
    def __read_title(session: Session):
        session_response = requests.get(
            url=f'{SESSION_URL}{session.session_id}',
            auth=SensoriaIoClient.token())

        if session_response.reason == 'OK':
            api_result = session_response.json()['APIResult']
            metadata = api_result['MetaData']
            user_id = int(api_result['UserId'])

            session.set_title(title=metadata['Title'])
            session.set_user_id(user_id=user_id)
        else:
            logging.warning(f'\tCan\'t retrieve title for session '
                            f'{session.session_id}. '
                            f'Reason: {session_response.reason}')

    def __load_sessions(self) -> Set[int]:
        self.clear()

        path = join(self.__sessions_dir, SESSIONS_FILENAME)

        if not os.path.exists(path):
            return set()

        found_ids = set()

        with open(join(self.__sessions_dir, SESSIONS_FILENAME), 'rb') as f:
            sessions = pickle.load(f)

        for session_id, session in sessions.items():
            self.update({session.session_id: session})
            found_ids.add(session_id)

        return found_ids


class TokenRetrievingFailed(Exception):
    def __init__(self, message: str):
        super().__init__(message)



#  tools.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import numpy as np
import pandas as pd

__all__ = ['get_nans', 'get_single_valued_tsfresh']


def get_nans(df: pd.DataFrame, threshold: float = None):
    """
    :param df: pandas.DataFrame
    :param threshold: value in [0, 1] = percentage of nans to consider
    :return:
    """
    if threshold is None:
        threshold = 0.0

    nans_df = df.isna()
    nans_mean = nans_df.mean()
    nans_sum = nans_df.sum()

    nans = pd.concat({'mean': nans_mean,
                      'sum': nans_sum}, axis=1)

    nans = nans[nans['mean'] > threshold].sort_values('mean')

    return nans


def get_single_valued_tsfresh(db: pd.DataFrame):
    """
    Returns the columns containing just one value grouped by original feature.
    Useful to detect those tsfresh features which have been calculated in the
    wrong manner.
    :param db:
    :return:
    """

    uniques = db.nunique()
    uniques = uniques[uniques == 1]
    uniques.index = pd.MultiIndex.from_tuples(
        [x.split('__', 1) for x in uniques.index])

    uniques = uniques.unstack().T.replace(1.0, 'x').replace(np.nan, '')

    return uniques



#  session.py
#  Project: sensorialytics
#  Copyright (c) 2021 Sensoria Health Inc.
#  All rights reserved

import logging
import os
from typing import Dict, Tuple, Set

from sensorialytics.helpers.tools import as_bool
from . import helper as h
from .core_data import CoreData
from .events import Events
from .helper import FiltersType

__all__ = ['Session']


class Session(Dict[str, CoreData]):
    CORE_DATA_CLASS = CoreData

    def __init__(self, session_dir: str):
        """
        :param session_dir: directory of the session containing all the files
        """
        super().__init__()

        self._session_id = h.get_session_id(session_dir=session_dir)
        self._title = '** no title **'
        self._user_id = -1
        self.__core_index = 0
        self._tags = set()
        self._metadata = {}

        self._events = Events()
        self._aggregate = CoreData()

        self.__read_session_data(session_dir)

    def __str__(self):
        return f'{self.session_id} - {self.title}'

    @property
    def session_id(self) -> int:
        return self._session_id

    @property
    def user_id(self) -> int:
        return self._user_id

    @property
    def tags(self) -> Set:
        return self._tags

    @property
    def metadata(self):
        return self._metadata

    @property
    def title(self) -> str:
        """
        :return: title of the session
        """
        return self._title

    @property
    def aggregate(self) -> CoreData:
        """
        :return: aggregate data for the session
        """
        return self._aggregate

    @property
    def events(self) -> Events:
        """
        :return: events occurred during the session
        """
        return self._events

    @property
    def settings(self) -> dict:
        """
        :return: settings used during the session
        """
        return self._events.settings

    @property
    def leading_core(self) -> str:
        """
        :return: str leading core for the session
        """
        for core_data in self.values():
            if core_data.metadata.get(h.KEY_IS_LEADING_CORE, False):
                return core_data.core_name[-4:]

        return ""

    def __getitem__(self, item: str) -> CoreData:
        """
        :param item str = name of the core used or tag. Default tags are
                assigned as C0, C1, ...
        :return: CoreData
        """
        if self.keys().__contains__(item):
            return super(Session, self).__getitem__(item)

        for core_data in self.values():
            if item in core_data.tags:
                return core_data

        raise Exception(f'No CoreData with tag or id \'{item}\'')

    def add_tag(self, tag: str):
        self._tags.add(tag)

    def remove_tag(self, tag: str):
        self._tags.remove(tag)

    def clear_tags(self):
        self._tags.clear()

    def filter(self, filters: FiltersType) -> None:
        """
        Filters the raw data passing them through filters
        :param filters:
            can be:
                - dict with {column_name: Filter}
                    e.g. {'Ax': ButterworthFilter}
                - dict with {column_name: [Filter, Filter, ..]}
                    e.g. {'Ax': [
                                    ButterworthFilter(...),
                                    RecursiveAverageFilter(...)
                                ]}
                    where the filters are applied in sequence
        """

        for core_data in self.values():
            core_data.filter(filters=filters)

    def offset(self, offsets: Dict[str, Dict[str, float]]) -> None:
        """
        Offsets the data using the values in offsets
        :param offsets: dict composed as:
            {
                core_id : {
                    column_name : offset
                }
            }
        """

        for core_data in self.values():
            offset = offsets.get(core_data.core_name)

            if offset is not None:
                core_data.offset(offset=offset)

    def subsample(self, target_sampling_frequency=None,
                  decimation_factor: int = None):
        for core_data in self.values():
            core_data.subsample(
                target_sampling_frequency=target_sampling_frequency,
                decimation_factor=decimation_factor
            )

    def set_title(self, title: str) -> None:
        """
        Sets the title of the session
        :param title: title of the session
        """

        self._title = title.lower()

    def set_user_id(self, user_id: int) -> None:
        """
        Sets the title of the session
        :param user_id: title of the session
        """

        self._user_id = user_id

    def update_metadata(self, metadata: dict):
        self._metadata.update(metadata)

    def clear_metadata(self):
        self._metadata = {}

    def __read_session_data(self, session_dir: str):
        logging.info(f'[{self._session_id}]')

        file_paths = self.__get_files_paths(session_dir)
        organized_file_paths = self.__organize_file_paths(file_paths)

        self.__load_core_data(organized_file_paths=organized_file_paths)
        self.__load_events(organized_file_paths=organized_file_paths)
        self.__load_aggregate(organized_file_paths=organized_file_paths)

    def __organize_file_paths(self, file_paths):
        organized_file_paths = {}

        for file_path in file_paths:
            log_type, core_name = self.__get_log_type(file_path)

            if core_name is None:
                organized_file_paths.update({log_type: file_path})
            else:
                if h.LOG_TYPE_CORE_DATA not in organized_file_paths.keys():
                    organized_file_paths.update({h.LOG_TYPE_CORE_DATA: {}})

                core_data = organized_file_paths.get(h.LOG_TYPE_CORE_DATA)

                if core_name not in core_data.keys():
                    core_data.update({core_name: {}})

                core_data[core_name].update({log_type: file_path})

        return organized_file_paths

    def __load_core_data(self, organized_file_paths: dict):
        core_data_paths = organized_file_paths.get(h.LOG_TYPE_CORE_DATA, None)

        if core_data_paths is not None:
            for core_name, paths in core_data_paths.items():
                core_data = self.CORE_DATA_CLASS(
                    raw_data_path=paths.get(h.LOG_TYPE_RAW, None),
                    processed_data_path=paths.get(h.LOG_TYPE_PROCESSED, None)
                )

                core_data.add_tag(f'C{self.__core_index}')
                self.__core_index += 1

                self.update({core_name[-4:]: core_data})

    def __load_events(self, organized_file_paths: dict):
        events_path = organized_file_paths.get(h.LOG_TYPE_EVENTS)

        if events_path is not None:
            self._events = Events(events_path=events_path)

            leading_core_data = self.get(self.leading_core, None)

            if leading_core_data is not None:
                time = leading_core_data.processed[h.COL_TIME]
                self._events[h.COL_TIME] = time
                self._events[h.COL_TIME].fillna(0.0, inplace=True)

    def __load_aggregate(self, organized_file_paths: dict):
        aggregate_path = organized_file_paths.get(h.LOG_TYPE_AGGREGATE)

        if aggregate_path is not None:
            self._aggregate = self.CORE_DATA_CLASS(raw_data_path=aggregate_path)

    @staticmethod
    def __get_files_paths(session_dir):
        files_paths = [
            os.path.join(session_dir, core_data_filename)
            for core_data_filename in os.listdir(session_dir)
            if h.is_session_name(core_data_filename)
        ]

        return list(sorted(files_paths))

    @staticmethod
    def __get_log_type(path) -> Tuple[str, str]:
        log_type = h.LOG_TYPE_RAW
        core_name = None

        with open(path, 'r') as f:
            lines = f.readlines()[0:50]

        for line in lines:
            if h.KEY_CSV_LOG_TYPE + ':' in line:
                log_type = line.split(h.METADATA_SEPARATOR)[-1].strip()
            if h.KEY_CSV_DEVICE_NAME in line:
                core_name = line.split(h.METADATA_SEPARATOR)[-1].strip()

        return log_type, core_name

    @staticmethod
    def __take(x, i: int):
        try:
            return x[i].strip()
        except IndexError:
            return None

    def __parse_event_info(self, x):
        if x is None:
            return None

        all_info = self.__take(x.split(':'), 1)

        if all_info is None:
            return None

        parsed_event_info = {}

        for info in all_info.split('|'):
            info_kv = info.strip().split('=')
            k = self.__parse_value(info_kv[0].strip())
            v = self.__parse_value(info_kv[1].strip())

            parsed_event_info.update({k: v})

        return parsed_event_info

    @staticmethod
    def __parse_value(x):
        if x is None:
            return None

        if x in ['true', 'false', 'True', 'False']:
            return as_bool(x)

        try:
            return float(x)
        except ValueError:
            return x
